@@ -28,6 +28,13 @@
 #include "tvgMath.h"
 #include "tvgLines.h"
 #include "tvgLottieInterpolator.h"
+#include "tvgLottieExpressions.h"
+
+
+struct LottieFont;
+struct LottieLayer;
+struct LottieObject;
+
 
 struct PathSet
 {
@@ -51,8 +58,6 @@ struct ColorStop
 };
 
 
-struct LottieFont;
-
 struct TextDocument
 {
     char* text = nullptr;
@@ -70,8 +75,8 @@ struct TextDocument
     } stroke;
     char* name = nullptr;
     float size;
+    float tracking = 0.0f;
     uint8_t justify;
-    uint8_t tracking;
 };
 
 
@@ -93,35 +98,6 @@ static inline RGB24 operator*(const RGB24& lhs, float rhs)
 }
 
 
-static void copy(PathSet& pathset, Array<Point>& outPts, Matrix* transform)
-{
-    Array<Point> inPts;
-
-    if (transform) {
-        for (int i = 0; i < pathset.ptsCnt; ++i) {
-            Point pt = pathset.pts[i];
-            mathMultiply(&pt, transform);
-            outPts.push(pt);
-        }
-    } else {
-        inPts.data = pathset.pts;
-        inPts.count = pathset.ptsCnt;
-        outPts.push(inPts);
-        inPts.data = nullptr;
-    }
-}
-
-
-static void copy(PathSet& pathset, Array<PathCommand>& outCmds)
-{
-    Array<PathCommand> inCmds;
-    inCmds.data = pathset.cmds;
-    inCmds.count = pathset.cmdsCnt;
-    outCmds.push(inCmds);
-    inCmds.data = nullptr;
-}
-
-
 template<typename T>
 struct LottieScalarFrame
 {
@@ -192,31 +168,136 @@ struct LottieVectorFrame
 };
 
 
-template<typename T>
-uint32_t bsearch(T* frames, float frameNo)
-{
-    uint32_t low = 0;
-    uint32_t high = frames->count - 1;
-
-    while (low <= high) {
-        auto mid = low + (high - low) / 2;
-        auto frame = frames->data + mid;
-        if (mathEqual(frameNo, frame->no)) return mid;
-        else if (frameNo < frame->no) high = mid - 1;
-        else low = mid + 1;
-    }
-    if (high < low) low = high;
-    return low;
-}
-
-
+//Property would have an either keyframes or single value.
 struct LottieProperty
 {
     enum class Type : uint8_t { Point = 0, Float, Opacity, Color, PathSet, ColorStop, Position, TextDoc, Invalid };
     virtual ~LottieProperty() {}
+
+    LottieExpression* exp = nullptr;
+
+    //TODO: Apply common bodies?
+    virtual uint32_t frameCnt() = 0;
+    virtual uint32_t nearest(float time) = 0;
+    virtual float frameNo(int32_t key) = 0;
 };
 
 
+struct LottieExpression
+{
+    enum LoopMode : uint8_t { None = 0, InCycle = 1, InPingPong, InOffset, InContinue, OutCycle, OutPingPong, OutOffset, OutContinue };
+
+    char* code;
+    LottieComposition* comp;
+    LottieLayer* layer;
+    LottieObject* object;
+    LottieProperty* property;
+    LottieProperty::Type type;
+
+    bool enabled;
+
+    struct {
+        uint32_t key = 0;      //the keyframe number repeating to
+        float in = FLT_MAX;    //looping duration in frame number
+        LoopMode mode = None;
+    } loop;
+;
+    ~LottieExpression()
+    {
+        free(code);
+    }
+};
+
+
+static void _copy(PathSet& pathset, Array<Point>& outPts, Matrix* transform)
+{
+    Array<Point> inPts;
+
+    if (transform) {
+        for (int i = 0; i < pathset.ptsCnt; ++i) {
+            Point pt = pathset.pts[i];
+            mathMultiply(&pt, transform);
+            outPts.push(pt);
+        }
+    } else {
+        inPts.data = pathset.pts;
+        inPts.count = pathset.ptsCnt;
+        outPts.push(inPts);
+        inPts.data = nullptr;
+    }
+}
+
+
+static void _copy(PathSet& pathset, Array<PathCommand>& outCmds)
+{
+    Array<PathCommand> inCmds;
+    inCmds.data = pathset.cmds;
+    inCmds.count = pathset.cmdsCnt;
+    outCmds.push(inCmds);
+    inCmds.data = nullptr;
+}
+
+
+template<typename T>
+uint32_t _bsearch(T* frames, float frameNo)
+{
+    int32_t low = 0;
+    int32_t high = int32_t(frames->count) - 1;
+
+    while (low <= high) {
+        auto mid = low + (high - low) / 2;
+        auto frame = frames->data + mid;
+        if (frameNo < frame->no) high = mid - 1;
+        else low = mid + 1;
+    }
+    if (high < low) low = high;
+    if (low < 0) low = 0;
+    return low;
+}
+
+
+template<typename T>
+uint32_t _nearest(T* frames, float frameNo)
+{
+    if (frames) {
+        auto key = _bsearch(frames, frameNo);
+        if (key == frames->count - 1) return key;
+        return (fabsf(frames->data[key].no - frameNo) < fabsf(frames->data[key + 1].no - frameNo)) ? key : (key + 1);
+    }
+    return 0;
+}
+
+
+template<typename T>
+float _frameNo(T* frames, int32_t key)
+{
+    if (!frames) return 0.0f;
+    if (key < 0) key = 0;
+    if (key >= (int32_t) frames->count) key = (int32_t)(frames->count - 1);
+    return (*frames)[key].no;
+}
+
+
+template<typename T>
+float _loop(T* frames, float frameNo, LottieExpression* exp)
+{
+    if (frameNo >= exp->loop.in || frameNo < frames->first().no ||frameNo < frames->last().no) return frameNo;
+
+    switch (exp->loop.mode) {
+        case LottieExpression::LoopMode::InCycle: {
+            frameNo -= frames->first().no;
+            return fmodf(frameNo, frames->last().no - frames->first().no) + (*frames)[exp->loop.key].no;
+        }
+        case LottieExpression::LoopMode::OutCycle: {
+            frameNo -= frames->first().no;
+            return fmodf(frameNo, (*frames)[frames->count - 1 - exp->loop.key].no - frames->first().no) + frames->first().no;
+        }
+        default: break;
+    }
+    return frameNo;
+}
+
+
 template<typename T>
 struct LottieGenericProperty : LottieProperty
 {
@@ -235,6 +316,26 @@ struct LottieGenericProperty : LottieProperty
     void release()
     {
         delete(frames);
+        frames = nullptr;
+        if (exp) {
+            delete(exp);
+            exp = nullptr;
+        }
+    }
+
+    uint32_t nearest(float frameNo) override
+    {
+        return _nearest(frames, frameNo);
+    }
+
+    uint32_t frameCnt() override
+    {
+        return frames ? frames->count : 1;
+    }
+
+    float frameNo(int32_t key) override
+    {
+        return _frameNo(frames, key);
     }
 
     LottieScalarFrame<T>& newFrame()
@@ -260,11 +361,21 @@ struct LottieGenericProperty : LottieProperty
         if (frames->count == 1 || frameNo <= frames->first().no) return frames->first().value;
         if (frameNo >= frames->last().no) return frames->last().value;
 
-        auto frame = frames->data + bsearch(frames, frameNo);
+        auto frame = frames->data + _bsearch(frames, frameNo);
         if (frame->no == frameNo) return frame->value;
         return frame->interpolate(frame + 1, frameNo);
     }
 
+    T operator()(float frameNo, LottieExpressions* exps)
+    {
+        T out{};
+        if (exps && (exp && exp->enabled)) {
+            if (exp->loop.mode != LottieExpression::LoopMode::None) frameNo = _loop(frames, frameNo, exp);
+            if (exps->result<LottieGenericProperty<T>>(frameNo, out, exp)) return out;
+        }
+        return operator()(frameNo);
+    }
+
     T& operator=(const T& other)
     {
         //shallow copy, used for slot overriding
@@ -292,6 +403,11 @@ struct LottiePathSet : LottieProperty
 
     void release()
     {
+        if (exp) {
+            delete(exp);
+            exp = nullptr;
+        }
+
         free(value.cmds);
         free(value.pts);
 
@@ -305,6 +421,21 @@ struct LottiePathSet : LottieProperty
         free(frames);
     }
 
+    uint32_t nearest(float frameNo) override
+    {
+        return _nearest(frames, frameNo);
+    }
+
+    uint32_t frameCnt() override
+    {
+        return frames ? frames->count : 1;
+    }
+
+    float frameNo(int32_t key) override
+    {
+        return _frameNo(frames, key);
+    }
+
     LottieScalarFrame<PathSet>& newFrame()
     {
         if (!frames) {
@@ -324,43 +455,43 @@ struct LottiePathSet : LottieProperty
         return (*frames)[frames->count];
     }
 
-    bool operator()(float frameNo, Array<PathCommand>& cmds, Array<Point>& pts, Matrix* transform = nullptr)
+    bool operator()(float frameNo, Array<PathCommand>& cmds, Array<Point>& pts, Matrix* transform)
     {
         if (!frames) {
-            copy(value, cmds);
-            copy(value, pts, transform);
+            _copy(value, cmds);
+            _copy(value, pts, transform);
             return true;
         }
 
         if (frames->count == 1 || frameNo <= frames->first().no) {
-            copy(frames->first().value, cmds);
-            copy(frames->first().value, pts, transform);
+            _copy(frames->first().value, cmds);
+            _copy(frames->first().value, pts, transform);
             return true;
         }
 
         if (frameNo >= frames->last().no) {
-            copy(frames->last().value, cmds);
-            copy(frames->last().value, pts, transform);
+            _copy(frames->last().value, cmds);
+            _copy(frames->last().value, pts, transform);
             return true;
         }
 
-        auto frame = frames->data + bsearch(frames, frameNo);
+        auto frame = frames->data + _bsearch(frames, frameNo);
 
         if (frame->no == frameNo) {
-            copy(frame->value, cmds);
-            copy(frame->value, pts, transform);
+            _copy(frame->value, cmds);
+            _copy(frame->value, pts, transform);
             return true;
         }
 
         //interpolate
-        copy(frame->value, cmds);
+        _copy(frame->value, cmds);
 
         auto t = (frameNo - frame->no) / ((frame + 1)->no - frame->no);
         if (frame->interpolator) t = frame->interpolator->progress(t);
 
         if (frame->hold) {
-            if (t < 1.0f) copy(frame->value, pts, transform);
-            else copy((frame + 1)->value, pts, transform);
+            if (t < 1.0f) _copy(frame->value, pts, transform);
+            else _copy((frame + 1)->value, pts, transform);
             return true;
         }
 
@@ -375,6 +506,16 @@ struct LottiePathSet : LottieProperty
         return true;
     }
 
+
+    bool operator()(float frameNo, Array<PathCommand>& cmds, Array<Point>& pts, Matrix* transform, LottieExpressions* exps)
+    {
+        if (exps && (exp && exp->enabled)) {
+            if (exp->loop.mode != LottieExpression::LoopMode::None) frameNo = _loop(frames, frameNo, exp);
+            if (exps->result<LottiePathSet>(frameNo, cmds, pts, transform, exp)) return true;
+        }
+        return operator()(frameNo, cmds, pts, transform);
+    }
+
     void prepare() {}
 };
 
@@ -393,6 +534,11 @@ struct LottieColorStop : LottieProperty
 
     void release()
     {
+        if (exp) {
+            delete(exp);
+            exp = nullptr;
+        }
+
         if (value.data) {
             free(value.data);
             value.data = nullptr;
@@ -408,6 +554,21 @@ struct LottieColorStop : LottieProperty
         frames = nullptr;
     }
 
+    uint32_t nearest(float frameNo) override
+    {
+        return _nearest(frames, frameNo);
+    }
+
+    uint32_t frameCnt() override
+    {
+        return frames ? frames->count : 1;
+    }
+
+    float frameNo(int32_t key) override
+    {
+        return _frameNo(frames, key);
+    }
+
     LottieScalarFrame<ColorStop>& newFrame()
     {
         if (!frames) {
@@ -427,27 +588,26 @@ struct LottieColorStop : LottieProperty
         return (*frames)[frames->count];
     }
 
-    void operator()(float frameNo, Fill* fill)
+    Result operator()(float frameNo, Fill* fill, LottieExpressions* exps)
     {
-        if (!frames) {
-            fill->colorStops(value.data, count);
-            return;
+        if (exps && (exp && exp->enabled)) {
+            if (exp->loop.mode != LottieExpression::LoopMode::None) frameNo = _loop(frames, frameNo, exp);
+            if (exps->result<LottieColorStop>(frameNo, fill, exp)) return Result::Success;
         }
 
+        if (!frames) return fill->colorStops(value.data, count);
+
         if (frames->count == 1 || frameNo <= frames->first().no) {
-            fill->colorStops(frames->first().value.data, count);
-            return;
+            return fill->colorStops(frames->first().value.data, count);
         }
 
         if (frameNo >= frames->last().no) {
-            fill->colorStops(frames->last().value.data, count);
-            return;
+            return fill->colorStops(frames->last().value.data, count);
         }
 
-        auto frame = frames->data + bsearch(frames, frameNo);
+        auto frame = frames->data + _bsearch(frames, frameNo);
         if (frame->no == frameNo) {
-            fill->colorStops(frame->value.data, count);
-            return;
+            return fill->colorStops(frame->value.data, count);
         }
 
         //interpolate
@@ -472,7 +632,7 @@ struct LottieColorStop : LottieProperty
             auto a = mathLerp(s->a, e->a, t);
             result.push({offset, r, g, b, a});
         }
-        fill->colorStops(result.data, count);
+        return fill->colorStops(result.data, count);
     }
 
     LottieColorStop& operator=(const LottieColorStop& other)
@@ -512,6 +672,27 @@ struct LottiePosition : LottieProperty
     void release()
     {
         delete(frames);
+        frames = nullptr;
+
+        if (exp) {
+            delete(exp);
+            exp = nullptr;
+        }
+    }
+
+    uint32_t nearest(float frameNo) override
+    {
+        return _nearest(frames, frameNo);
+    }
+
+    uint32_t frameCnt() override
+    {
+        return frames ? frames->count : 1;
+    }
+
+    float frameNo(int32_t key) override
+    {
+        return _frameNo(frames, key);
     }
 
     LottieVectorFrame<Point>& newFrame()
@@ -537,18 +718,28 @@ struct LottiePosition : LottieProperty
         if (frames->count == 1 || frameNo <= frames->first().no) return frames->first().value;
         if (frameNo >= frames->last().no) return frames->last().value;
 
-        auto frame = frames->data + bsearch(frames, frameNo);
+        auto frame = frames->data + _bsearch(frames, frameNo);
         if (frame->no == frameNo) return frame->value;
         return frame->interpolate(frame + 1, frameNo);
     }
 
+    Point operator()(float frameNo, LottieExpressions* exps)
+    {
+        Point out{};
+        if (exps && (exp && exp->enabled)) {
+            if (exp->loop.mode != LottieExpression::LoopMode::None) frameNo = _loop(frames, frameNo, exp);
+            if (exps->result<LottiePosition>(frameNo, out, exp)) return out;
+        }
+        return operator()(frameNo);
+    }
+
     float angle(float frameNo)
     {
         if (!frames) return 0;
         if (frames->count == 1 || frameNo <= frames->first().no) return 0;
         if (frameNo >= frames->last().no) return 0;
 
-        auto frame = frames->data + bsearch(frames, frameNo);
+        auto frame = frames->data + _bsearch(frames, frameNo);
         return frame->angle(frame + 1, frameNo);
     }
 
@@ -574,6 +765,11 @@ struct LottieTextDoc : LottieProperty
 
     void release()
     {
+        if (exp) {
+            delete(exp);
+            exp = nullptr;
+        }
+
         if (value.text) {
             free(value.text);
             value.text = nullptr;
@@ -593,6 +789,21 @@ struct LottieTextDoc : LottieProperty
         frames = nullptr;
     }
 
+    uint32_t nearest(float frameNo) override
+    {
+        return _nearest(frames, frameNo);
+    }
+
+    uint32_t frameCnt() override
+    {
+        return frames ? frames->count : 1;
+    }
+
+    float frameNo(int32_t key) override
+    {
+        return _frameNo(frames, key);
+    }
+
     LottieScalarFrame<TextDocument>& newFrame()
     {
         if (!frames) frames = new Array<LottieScalarFrame<TextDocument>>;
@@ -616,7 +827,7 @@ struct LottieTextDoc : LottieProperty
         if (frames->count == 1 || frameNo <= frames->first().no) return frames->first().value;
         if (frameNo >= frames->last().no) return frames->last().value;
 
-        auto frame = frames->data + bsearch(frames, frameNo);
+        auto frame = frames->data + _bsearch(frames, frameNo);
         return frame->value;
     }
 
